#include <easy.h>

e_obj_descriptor e_obj_descriptor_aset = {
  "aset",             /* Name      */
  0,                  /* def_mode  */
  malloc,             /* _alloc    */
  free,               /* _free     */
  realloc,            /* _realloc  */
  e_aset_free,        /* free      */
  e_aset_to_string,   /* to_string */
  e_aset_clone,       /* clone     */
  e_aset_equals,      /* equals    */
};

e_obj_descriptor e_obj_descriptor_aset_item = {
  "aset_item",        /* Name      */
  0,                  /* def_mode  */
  malloc,             /* _alloc    */
  free,               /* _free     */
  realloc,            /* _realloc  */
  e_aset_item_free,   /* free      */
  e_aset_item_to_string, /* to_string */
  e_aset_item_clone,  /* clone     */
  e_aset_item_equals, /* equals    */
};

/************************************************************/
/* Aset :                                                   */
/* Binary tree or aset                                      */
/************************************************************/
/***************************************************************************/
/* Create or initialise a new e_aset                                       */
/* asp : e_aset to initialise (static e_aset)                              */
/*      if NULL allocate it                                                */
/*      asp is assumed to be of the correct type                           */
/* Return e_aset                                                           */
/*        0 : error                                                        */
/* Link : return : +1                                                      */
/***************************************************************************/
e_aset *e_aset_new(e_aset *asp)
{ 
                 /* Static object or allocate it ? */
  if (!asp)
  { if (!(asp = e_obj_descriptor_aset._alloc(sizeof(e_aset)))) 
      return((e_aset *)e_error_no_memory("e_aset_new"));
    asp->h.refs = 1;
  }
  else asp->h.refs = -1;
  asp->h.mode = 0;

  asp->h.obj_desc_p = &e_obj_descriptor_aset;

  E_ASET_FIRST(asp) = 0;  
  E_ASET_LAST(asp) = 0;  
  E_ASET_CUR(asp) = 0;  
  E_ASET_CARDINAL(asp) = 0;

  return(asp);
}

/***************************************************************************/
/* Create or initialise a new e_aset_item                                  */
/* aip : e_aset_item to initialise (static e_aset)                         */
/*      if NULL allocate it                                                */
/*      aip is assumed to be of the correct type                           */
/* np  : name for the item (can be null)                                   */
/* val : value for the item (can be null)                                  */
/* Return e_aset_item                                                      */
/*        0 : error                                                        */
/* Link : np : -1                                                          */
/*        val : -1                                                         */
/*        Return : +1                                                      */
/***************************************************************************/
e_aset_item *e_aset_item_new(e_aset_item *aip,e_name *np,void *val)
{ 
                 /* Static object or allocate it ? */
  if (!aip)
  { if (!(aip = e_obj_descriptor_aset._alloc(sizeof(e_aset_item)))) 
      return((e_aset_item *)e_error_no_memory("e_aset_item_new"));
    aip->h.refs = 1;
  }
  else aip->h.refs = -1;
  aip->h.mode = 0;

  aip->h.obj_desc_p = &e_obj_descriptor_aset_item;

  E_ASET_ITEM_PREV(aip) = 0;
  E_ASET_ITEM_NEXT(aip) = 0;
  E_ASET_ITEM_NAME(aip) = np;  
  E_ASET_ITEM_VAL(aip) = val;  

  return(aip);
}

/***************************************************************************/
/* Free the e_aset                                                         */
/* If the refs counter is -1 : static aset : leave as is                   */
/* If the refs counter is 1 : free the allocated memory for the e_aset     */
/* Otherwise decrement the counter by 1.                                   */
/* REF : stp : -1                                                          */
/***************************************************************************/
void e_aset_free(void *asp)
{ e_aset_item *aip,*next;

  if(asp)
  { if(E_OBJ_REF(asp) == 0)
    { e_error(E_ERROR_INCONSISTENCY,E_ERROR_LEVEL_PANIC,
            "Trying to free a free aset (%x) ",asp);
      goto error;
    }
                                /* static aset */
    if (E_OBJ_REF(asp) < 0) return;

                                /* decrement the refs counter   */
    if (E_OBJ(asp)->refs-- > 1) return;

                                /* Last reference,     
                                   Free the two element, 
                                    then free the aset */
                 /* Free each item                        */
                 /* items can't be in more than one list. */
                 /* Their Ref counter should always be 1  */
    aip = E_ASET_FIRST(asp);
    while(aip)
    { next = E_ASET_ITEM_NEXT(aip);
      e_aset_item_free(aip);
      aip = next;
    }

    e_obj_descriptor_aset._free(asp);
  }
  return;

error:
  e_error_flush(E_ERROR_LEVEL_DEBUG); 
  
}

/***************************************************************************/
/* Free the e_aset_item                                                    */
/* e_aset_item  can only be inserted into 1 list, so their Ref counter     */
/* Must be one. This should only be called by functions manipulating       */
/* asets. They must deal with the pointers.                                */
/* The Name and value of the aset_item are freed (and the the e_aset_item) */
/* REF : aip : -1                                                          */
/***************************************************************************/
void e_aset_item_free(void *aip)
{ int free = 1;

  if(aip)
  { if(E_OBJ_REF(aip) == 0)
    { e_error(E_ERROR_INCONSISTENCY,E_ERROR_LEVEL_PANIC,
            "Trying to free a free aset_item (%x) ",aip);
      goto error;
    }
                                /* static aset */
    if (E_OBJ_REF(aip) < 0) 
    { free = 0; goto end;
    }
                                /* decrement the refs counter   */
    if (E_OBJ(aip)->refs != 1) 
    { e_error(E_ERROR_INCONSISTENCY,E_ERROR_LEVEL_PANIC,
            "aset_item has ref counter larger than 1 : (%x) ",aip);
      goto error;
    }

end:
                   /* Free name and value */
    if(E_ASET_ITEM_NAME(aip)) E_FREE(E_ASET_ITEM_NAME(aip));
    if(E_ASET_ITEM_VAL(aip)) E_FREE(E_ASET_ITEM_VAL(aip));

    if(free) e_obj_descriptor_aset_item._free(aip);
  }
  return;

error:
  e_error_flush(E_ERROR_LEVEL_DEBUG); 
  
}

/*************************************************************/
/* Convert the aset to a string                              */
/* aip : an e_aset, assumed to be of the correct type        */
/* { Name1 = Value;                                          */
/*   Name2 = Value1;                                         */
/*   ...                                                     */
/* }                                                         */
/* Return : new string                                       */
/*          0 : error                                        */
/* LINK : asp : 0                                            */
/*        RETURN : +1                                        */
/*************************************************************/
void *e_aset_to_string(void *asp,int tab,e_flag mode)
{ e_string *stp = 0,*stp2 = 0;
  e_aset_item *aip = 0;
  char *s;
                              /* write { before the first element */

  if(!E_ASET_CARDINAL(asp))  /* Empty set : {} */
  { E_NULL_ERROR(stp = e_string_new(0,"{ }",tab+5));

    if(!(mode & E_TO_STR_NO_TAB))
    { stp = e_string_insert_chars(stp,0,tab,' '); /* add tab to the string */
    }

    return(stp);
  }

  aip = E_ASET_FIRST(asp);
  E_NULL_ERROR(stp = e_string_new(0,"{ ",E_LINE_SIZE));
                             /* First item : same line as { */

  E_NULL_ERROR(stp = e_string_append_string(stp,(e_string *)
             e_aset_item_to_string(aip,tab+2,mode|E_TO_STR_NO_TAB_ON_FIRST))
              ); 

  if(mode & E_TO_STR_COMPACT) s = ";";
  else                        s = ";\n";
  E_NULL_ERROR(stp = e_string_append_buffer(stp,s));

  if(!(mode & E_TO_STR_NO_TAB))    
  { stp = e_string_insert_chars(stp,0,tab,' ');   /* add tab to the string */
  }

  while((aip = aip->next))
  { E_NULL_ERROR(stp2 = E_TO_STRING(aip,tab+2,
                                    mode&(~E_TO_STR_NO_TAB_ON_FIRST)));
    if(mode & E_TO_STR_COMPACT) s = ";";
    else                        s = ";\n";
    E_NULL_ERROR(stp2 = e_string_append_buffer(stp2,s));
    if(!(stp = e_string_append_string(stp,stp2))) goto error2;
  }

  E_NULL_ERROR(stp2 = e_string_new(0,"}",3));

  if(!(mode & E_TO_STR_COMPACT))
  { stp2 = e_string_insert_chars(stp2,0,tab,' ');   /* add tab to the string */
  }
  if(!(stp = e_string_append_string(stp,stp2))) goto error2;

  return (stp);

error:
  if (stp2) E_FREE(stp2);
error2:
  if (stp) E_FREE(stp);
  return(0);
}

/*************************************************************/
/* Convert the aset_item to a string                         */
/* Name = Value or Value                                     */
/* Return : new string                                       */
/*          0 : error                                        */
/* LINK : aip : 0                                            */
/*        RETURN : +1                                        */
/*************************************************************/
void *e_aset_item_to_string(void *aip,int tab,e_flag mode)
{ e_string *stp = 0,*stp2 = 0;
  
                                    /* Name = */
  if(E_ASET_ITEM_NAME(aip))
  { E_NULL_ERROR(stp = E_TO_STRING(E_ASET_ITEM_NAME(aip),0,mode));
    E_NULL_ERROR(stp = e_string_append_buffer(stp," = "));
  }
                                    /* ... Value */
  if(E_ASET_ITEM_VAL(aip))
  { E_NULL_ERROR(stp2 = E_TO_STRING(E_ASET_ITEM_VAL(aip),tab,
                                    mode|E_TO_STR_NO_TAB_ON_FIRST));
  }

  if(stp) { E_NULL_ERROR(stp = e_string_append_string(stp,stp2)); }
  else stp = stp2;

  if(!(mode & E_TO_STR_NO_TAB))
  { stp = e_string_insert_chars(stp,0,tab,' '); /* add tab to the string */
  }
  return(stp);

error:
  if (stp) E_FREE(stp);
  return(0);
}

/*************************************************************************/
/* Clone an aset                                                         */
/* mode : cloning mode                                                   */
/* If the aset can't be cloned, return asp and increase its refs counter */
/* asp : an aset, assumed to be of the correct type                      */
/* mode : to clone the object recursively                                */
/* Return : cloned e_aset                                                */
/*          0 : Error                                                    */
/* LINK : asp : 0                                                        */
/*        RETURN : +1                                                    */
/*************************************************************************/
void *e_aset_clone(void *asp,e_flag mode)
{ e_aset *new_asp=0;
  e_aset_item *aip,*new_aip,*aip_last=0;

                 /* Can't clone : return the same object */
  if(E_OBJ_NO_CLONE(asp)) return(E_USE(asp));

  E_NULL_ERROR(new_asp = e_aset_new(0));

              /* The aset is always cloned recursively */
  aip = E_ASET_FIRST(asp);
  while(aip)
  { E_NULL_ERROR(new_aip = e_aset_item_clone(aip,mode));
    if(!aip_last) E_ASET_FIRST(new_asp) = new_aip;
    else E_ASET_ITEM_NEXT(aip_last) = new_aip;
    E_ASET_ITEM_PREV(new_aip) = aip_last;

              /* Clone cur item pointer */
    if(aip == E_ASET_CUR(asp)) E_ASET_CUR(new_asp) = new_aip;

    aip_last = aip;
    aip = E_ASET_ITEM_NEXT(aip);
  }
  E_ASET_LAST(new_asp) = aip_last;
  E_ASET_CARDINAL(new_asp) = E_ASET_CARDINAL(asp);

  return(new_asp);

error:
  if(new_asp)
  { if((aip = E_ASET_FIRST(new_asp)))
    { while(aip) 
      { new_aip = E_ASET_ITEM_NEXT(aip);
        e_aset_item_free(aip);
        aip = new_aip;
      }
    } 
    e_obj_descriptor_aset_item._free(new_asp);
  }  
  return(0);
}

/*************************************************************************/
/* Clone an aset_item                                                    */
/* mode : cloning mode                                                   */
/* The next and prev pointers are not initialised, only the names and    */
/* the values are cloned.                                                */
/* aip : an aset_item, assumed to be of the correct type                 */
/* mode : to clone the object recursively                                */
/* Return : cloned e_aset                                                */
/*          0 : Error                                                    */
/* LINK : asp : 0                                                        */
/*        RETURN : +1                                                    */
/*************************************************************************/
void *e_aset_item_clone(void *aip,e_flag mode)
{ e_aset_item *new_aip;
  e_name *np;
  void *val;

                 /* Can't clone : return the same object */
  if(E_OBJ_NO_CLONE(aip)) return(E_USE(aip));

  if((np = E_ASET_ITEM_NAME(aip))) E_OBJ_REF(np)++;  
  val = E_ASET_ITEM_VAL(aip);  

                 /* Clone val or increase counter ? */
  if(mode & E_CLONE_RECURSIVE) val = E_CLONE(val,mode);
  else                         val = E_USE(val);

  E_NULL_ERROR(new_aip = e_aset_item_new(0,np,val));

  return(new_aip);

error:
  if(val && (mode & E_CLONE_RECURSIVE)) E_FREE(val);
  if(np) E_FREE(np);
  return(0);
}

/***************************************************************/
/* Test if the 2 asets are equal:                              */
/* The two asets are equal if they are identical (p1 == p2),   */
/* If they have the same cardinal and identical elements.      */
/* The current item is not checked!                            */
/* Return : 1 : p1 and p2 are equal                            */
/*          0 : p1 and p2 are different                        */
/* Link : p1 0                                                 */
/*        p2 0                                                 */
/***************************************************************/
int e_aset_equals(void *p1,void *p2)
{ e_aset_item *aip1,*aip2;

  if (p1 == p2)   return(1);
  if (!p1 || !p2) return(0);

            /* different objet  types => differ */
  if(E_OBJ_DESCR(p1) != E_OBJ_DESCR(p2)) return(0);

            /* compare the cradinals */
  if(E_ASET_CARDINAL(p1) != E_ASET_CARDINAL(p2)) return(0);

  aip1 = E_ASET_FIRST(p1);
  aip2 = E_ASET_FIRST(p2);
  while(aip1 && aip2)
  { if(!e_aset_item_equals(aip1,aip2)) return(0);
    aip1 = E_ASET_ITEM_NEXT(aip1);
    aip2 = E_ASET_ITEM_NEXT(aip2);
  }

  return(1);
}

/***************************************************************/
/* Test if the 2 aset_items are equal:                         */
/* The two asets_items are equal if they are identical (p1==p2)*/
/* or if they have the same value and the same name            */
/* Return : 1 : p1 and p2 are equal                            */
/*          0 : p1 and p2 are different                        */
/* Link : p1 0                                                 */
/*        p2 0                                                 */
/***************************************************************/
int e_aset_item_equals(void *p1,void *p2)
{ void *val1,*val2;

  if (p1 == p2)   return(1);
  if (!p1 || !p2) return(0);

  if(e_name_equals(E_ASET_ITEM_NAME(p1),E_ASET_ITEM_NAME(p2))) return(0);

  val1 = E_ASET_ITEM_VAL(p1);
  val2 = E_ASET_ITEM_VAL(p1);

  if(val1 == val2) return(1);  
  if(!val1) return(0);

  return(E_EQUALS(val1,val2));
}


/*********************************************************************/
/* Get the next item value from the aset and write it in obj.        */
/* asp : aset to scan : supposed to be of type e_aset.               */
/* asp.cur is - FIRST when refering to the first element.            */
/*            - 0 when the end of set has been reached.              */
/* mode : E_NEXT_GET_NEXT   : (0) get next item value                */
/*        E_NEXT_GET_CUR : get the current item value,               */
/*                         the counter is left unchanged.            */
/*        E_NEXT_GET_NEXT_KEY : get the next item key name           */
/*        E_NEXT_GET_CUR_KEY  : get the current item key name        */
/*                         the counter is left unchanged.            */
/*        E_NEXT_RESET : reset index counter to the first element    */
/* obj : pointer to the object pointer variable: in which the object */
/*       must be written into with a link                            */
/* Return : 1 : OK                                                   */
/*          -1 : no more elements                                    */
/*          0 : error or no more element                             */
/* Link : asp : 0                                                    */
/*        *obj : +1                                                  */
/*********************************************************************/
int e_aset_next(void *asp,int mode,void **obj)
{ 
  if(!asp) { e_error_null_pointer("e_aset_next(asp,-,-)"); goto error;}

  if(mode == E_NEXT_RESET) /* Reset the aset counter */
  { E_ASET_CUR(asp) = E_ASET_FIRST(asp);
    return(1);
  }

  if(!obj) { e_error_null_pointer("e_aset_next(-,-,obj)"); goto error;}

  if(E_ASET_CUR(asp) == 0) /* end of aset */
  { return(-1);
  }

  switch(mode)
  { case E_NEXT_GET_NEXT :
    case E_NEXT_GET_CUR :
           if((*obj = E_ASET_ITEM_VAL(E_ASET_CUR(asp)))) E_REF(*obj);
           break;
    case E_NEXT_GET_NEXT_KEY :
    case E_NEXT_GET_CUR_KEY :
           if((*obj = E_ASET_ITEM_NAME(E_ASET_CUR(asp)))) E_REF(*obj);
           break;
    default : 
           break;
  } 

  if((mode == E_NEXT_GET_NEXT)||(mode == E_NEXT_GET_NEXT_KEY))
  { E_ASET_CUR(asp) = E_ASET_ITEM_NEXT(E_ASET_CUR(asp));
  }

  return(1); 

error:
  
  return(0);
}

/*************************************************/
/* Add one item at the end of the aset asp.      */
/* If the aset is used more than 1, it is cloned */
/* and the clone is modified                     */
/* asp : aset to add the item to.                */
/* np : name for the item (can be NULL)          */
/* val : value for the new item (Can't be NULL)  */
/* Return : the modified aset.                   */
/*          0 : error                            */
/* Link : asp -1                                 */
/*        np  -1                                 */
/*        val -1                                 */
/*        Return : 1                             */
/*************************************************/
e_aset *e_aset_append_item(e_aset *asp,e_name *np, void *val)
{ e_aset *p;
  e_aset_item *aip = 0,*last;

  if(!asp) { e_error_null_pointer("e_aset_append_item(asp,-,-)"); goto error;}
  if(!val) { e_error_null_pointer("e_aset_append_item(-,-,val)"); goto error;}

             /* Make sure the aset is used only once */
  if(E_OBJ_REF(asp) > 1) 
  { p = E_CLONE(asp,0); E_FREE(asp);
    asp = p;
  }

             /* Allocate the new item */ 
  aip = e_aset_item_new(0,np,val);
  val = 0; np = 0; 
  E_NULL_ERROR(aip);

             /* Merge the item at the end of the list */
  if(!(last = E_ASET_LAST(asp)))
  {          /* The aset is empty => aip is the only element */
    E_ASET_LAST(asp) = aip;
    E_ASET_FIRST(asp) = aip;
    E_ASET_CARDINAL(asp) = 1;
  }
  else       /* add item to end of list */
  { E_ASET_ITEM_NEXT(last) = aip;
    E_ASET_ITEM_PREV(aip) = last;
    E_ASET_LAST(asp) = aip;
    E_ASET_CARDINAL(asp)++;
  }  
  return(asp);

error:
  if (asp) E_FREE(asp);  
  if (aip) E_FREE(aip);  
  if (np) E_FREE(np);  
  if (val) E_FREE(val);  
  return(0);
}

/*************************************************/
/* Add one item at the beginning of to the aset  */
/* If the aset is used more than 1, it is cloned */
/* and the clone is modified                     */
/* asp : aset to add the item to.                */
/* np : name for the item (can be NULL)          */
/* val : value for the new item (Can't be NULL)  */
/* Return : the modified aset.                   */
/*          0 : error                            */
/* Link : asp -1                                 */
/*        np  -1                                 */
/*        val -1                                 */
/*        Return : 1                             */
/*************************************************/
e_aset *e_aset_prepend_item(e_aset *asp,e_name *np, void *val)
{ e_aset *p;
  e_aset_item *aip = 0,*first;

  if(!asp) { e_error_null_pointer("e_aset_append_item(asp,-,-)"); goto error;}
  if(!val) { e_error_null_pointer("e_aset_append_item(-,-,val)"); goto error;}

             /* Make sure the aset is used only once */
  if(E_OBJ_REF(asp) > 1) 
  { p = E_CLONE(asp,0); E_FREE(asp);
    asp = p;
  }

             /* Allocate the new item */ 
  aip = e_aset_item_new(0,np,val);
  val = 0; np = 0; 
  E_NULL_ERROR(aip);

             /* Add the item at the beginning of the list */
  if(!(first = E_ASET_LAST(asp)))
  {          /* The aset is empty => aip is the only element */
    E_ASET_LAST(asp) = aip;
    E_ASET_FIRST(asp) = aip;
    E_ASET_CARDINAL(asp) = 1;
  }
  else       /* add item to end of list */
  { E_ASET_ITEM_PREV(first) = aip;
    E_ASET_ITEM_NEXT(aip) = first;
    E_ASET_FIRST(asp) = aip;
    E_ASET_CARDINAL(asp)++;
  }  
  return(asp);

error:
  if (asp) E_FREE(asp);  
  if (aip) E_FREE(aip);  
  if (np) E_FREE(np);  
  if (val) E_FREE(val);  
  return(0);
}

/*****************************************************************************/
/* Add one item to the aset asp. The item is normally added at the end       */
/* unless the select mode specifies otherwise.                               */
/* If the aset is used more than 1, it is cloned and the clone is modified   */
/* asp : aset to add the item to.                                            */
/* np : name for the item key (can be NULL)                                  */
/* val : value for the new item (Can't be NULL)                              */
/* mode : E_ASET_ADD_MODE_UNIQUE_VAL : add the object if it is not in the    */
/*      set yet (usualy the key wil be null for this but does not have to)   */
/*      If np is null the name of the object found is not changed.           */
/*      E_ASET_ADD_MODE_UNIQUE_KEY : add the object if the key is not in     */
/*      the set yet. If it is val replaces the old value unless the          */
/*      NO_REPLACE mode is selected too. when both UNIQUE_VAL and UNIQUE_KEY */
/*      are selected UNIQUE_VAL is ignored unless np is NULL.                */
/*      E_ASET_ADD_MODE_NO_REPLACE : when used with mode UNIQUE_KEY do not   */
/*      change the value for the key, and return the set as it was (no error)*/
/* Return : the modified aset.                                               */
/*          0 : error                                                        */
/* Link : asp -1                                                             */
/*        np  -1                                                             */
/*        val -1                                                             */
/*        Return : 1                                                         */
/*****************************************************************************/
/**
 * e_aset *e_aset_add_item(e_aset *asp,e_name *np, void *val,int mode)
 *                        [-1,-1,-1,-;1]
 **/
/*****************************************************************************/
e_aset *e_aset_add_item(e_aset *asp,e_name *np, void *val,int mode)
{ e_aset *p;
  e_aset_item *aip = 0;

  if(!asp) 
  { e_error_null_pointer("e_aset_append_item(asp,-,-,-)"); goto error;}
  if(!val) 
  { e_error_null_pointer("e_aset_append_item(-,-,val,-)"); goto error;}

             /* Make sure the aset is used only once */
  if(E_OBJ_REF(asp) > 1) 
  { p = E_CLONE(asp,0); E_FREE(asp);
    asp = p;
  }

                  /* Add item ensuring the key remains unique */
  if((mode & E_ASET_ADD_MODE_UNIQUE_KEY) && np)
  { 
    if(!(aip = e_aset_get_item_by_key(asp,np)))
    {                      /* Key not found : add  key = val  */
      return(e_aset_append_item(asp,np,val));
    }

    if(mode & E_ASET_ADD_MODE_NO_REPLACE)
    {                   /* Leave key value as it is */
      E_FREE(np); E_FREE(val);
      return(asp);
    }

    if(E_ASET_ITEM_VAL(aip)) E_FREE(E_ASET_ITEM_VAL(aip));
    E_ASET_ITEM_VAL(aip) = val;
    return(asp);
  }

                  /* Add item ensuring the value remains unique */
  if(mode & E_ASET_ADD_MODE_UNIQUE_VAL)
  { 
    if(!(aip = e_aset_get_item_by_key(asp,val)))
    {                      /* object not found : add  key = val  */
      return(e_aset_append_item(asp,np,val));
    }

    if(E_ASET_ITEM_VAL(aip)) E_FREE(E_ASET_ITEM_VAL(aip));
    E_ASET_ITEM_VAL(aip) = val;
    if(np)
    { if(E_ASET_ITEM_NAME(aip)) E_FREE(E_ASET_ITEM_NAME(aip));
      E_ASET_ITEM_NAME(aip) = np;
    }
    return(asp);
  }

                    /* By default simply append the item */
  return(e_aset_append_item(asp,np,val));

error:
  if (asp) E_FREE(asp);  
  if (np) E_FREE(np);  
  if (val) E_FREE(val);  
  return(0);
}

/*****************************************************************************/
/* Add a set to the aset asp. The items are normally added at the end        */
/* unless the select mode specifies otherwise.                               */
/* If the aset is used more than 1, it is cloned and the clone is modified   */
/* asp : aset to add the item to.                                            */
/* aspv : aset to add. Its items are added one by one                        */
/* mode : Same as for e_aset_add_item but it aplies for each element of aspv.*/
/* Return : the modified aset.                                               */
/*          0 : error                                                        */
/* Link : asp  -1                                                            */
/*        aspv -1                                                            */
/*        Return : 1                                                         */
/*****************************************************************************/
/**
 * e_aset *e_aset_add_aset(e_aset *asp,e_name *aspv,int mode)
 *                        [-1,-1,-;1]
 **/
/*****************************************************************************/
e_aset *e_aset_add_aset(e_aset *asp,e_aset *aspv,int mode)
{ e_aset_item *aip;    
  void *v;
  e_name *np;

  aip = E_ASET_FIRST(aspv);
  while(aip)
  { v = E_ASET_ITEM_VAL(aip);
    if((np = E_ASET_ITEM_NAME(aip))) E_USE(np);

    asp = e_aset_add_item(asp,np,E_USE(v),mode);
    aip = E_ASET_ITEM_NEXT(aip);
  }
  E_FREE(aspv);
  return(asp);
}

/*************************************************/
/* Append one aset at the end of asp.            */
/* If the aset is used more than 1, it is cloned */
/* and the clone is modified                     */
/* asp : aset to add the aset to.                */
/* asp2 : aset to append.                        */
/* Return : the modified aset.                   */
/*          0 : error                            */
/* Link : asp  -1                                */
/*        asp2 -1                                */
/*        Return : 1                             */
/*************************************************/
e_aset *e_aset_append(e_aset *asp,e_aset *asp2)
{ e_aset *p;
  e_aset_item *aip2,*p1,*p2;
  e_name *np;
  void *valp;

  if(!asp) { e_error_null_pointer("e_aset_append_item(asp,-)"); goto error;}
  if(!asp2) { e_error_null_pointer("e_aset_append_item(-,asp2)"); goto error;}

             /* Make sure the aset is used only once */
  if(E_OBJ_REF(asp) > 1) 
  { p = E_CLONE(asp,0); E_FREE(asp);
    asp = p;
  }

  if(E_OBJ_REF(asp2) == 1) 
  {          /* Use asp2 as it is, no need to copy each item */
    p1 = E_ASET_LAST(asp);
    p2 = E_ASET_FIRST(asp2);
    E_ASET_ITEM_NEXT(p1) = p2;
    E_ASET_ITEM_PREV(p2) = p1;
    E_ASET_LAST(asp) = E_ASET_LAST(asp2);
    E_ASET_CARDINAL(asp) += E_ASET_CARDINAL(asp2);

    E_ASET_CARDINAL(asp2) = 0;
    E_ASET_FIRST(asp2) = 0;
    E_ASET_LAST(asp2) = 0;
    goto end;
  }

             /* Else : Append each element one by one */
  aip2 = E_ASET_FIRST(asp2);

  while(aip2)
  { if((np = E_ASET_ITEM_NAME(aip2))) E_REF(np);
    if((valp = E_ASET_ITEM_VAL(aip2))) E_REF(valp);

    E_NULL_ERROR(asp = e_aset_append_item(asp,np,valp));

    aip2 = E_ASET_ITEM_NEXT(aip2);
  }

end:
  E_FREE(asp2);
  return(asp);

error:
  if (asp) E_FREE(asp);  
  if (asp2) E_FREE(asp2);  
  return(0);
}

/*********************************************************************/
/* Return the index of key in the aset staring at position start     */
/* The first element has index 0                                     */
/* Example :                                                         */    
/*  i = 0;                                                           */
/*  while((i = e_aset_get_key_index(asp, e_name_new("test"),i)) > 0) */
/*  { printf("index = %d\n",i++);                                    */
/*  }                                                                */
/* asp : aset to scan.                                               */
/* np : key name                                                     */
/* start : index of the first element to check                       */
/* Return : index position of the key in the set                     */
/*          -1 : key not found                                       */
/*          -2 : error                                               */
/* Link : asp  0                                                     */
/*        np  -1                                                     */
/*********************************************************************/
int e_aset_get_key_index(e_aset *asp, e_name *np,int start)
{ int i; 
  e_aset_item *aip;

  E_NULL_ERROR(e_check_type(asp,E_TYPE_ASET,"e_aset_get_key_index(asp,-,-)"));
  E_NULL_ERROR(e_check_type(np,E_TYPE_NAME,"e_aset_get_key_index(-,np,-)"));

  i = 0;
  aip = E_ASET_FIRST(asp);

                       /* Move to start */
  while((i < start) && aip)
  { aip = E_ASET_ITEM_NEXT(aip);
    i++;
  }
  
  while(aip)
  { if(E_ASET_ITEM_NAME(aip) == np) 
    { E_FREE(np);
      return(i);
    }
    aip = E_ASET_ITEM_NEXT(aip);
    i++;
  }
  E_FREE(np);
                       /* Key not found */
  return(-1);

error:
  if(np) E_FREE(np);
  return(-2);  
}

/*************************************************************************/
/* Return the index of the element obj in the aset staring at position   */
/* start. Return -1 if obj does not belong to the set.                   */
/* The first element has index 0                                         */
/* Example :                                                             */    
/*  i = 0;                                                               */
/*  while((i = e_aset_get_element_index(asp,val,i)) > 0)                 */
/*  { printf("index = %d\n",i++);                                        */
/*  }                                                                    */
/* asp : aset to scan.                                                   */
/* obj : object to find                                                  */
/* start : index of the first element to check                           */
/* Return : index position of the key in the set                         */
/*          -1 : key not found                                           */
/*          -2 : error                                                   */
/* Link : asp  0                                                         */
/*        obj  -1                                                        */
/*************************************************************************/
int e_aset_get_element_index(e_aset *asp,void *obj,int start)
{ int i; 
  e_aset_item *aip;

  E_NULL_ERROR(e_check_type(asp,E_TYPE_ASET,
                              "e_aset_get_element_index(asp,-,-)"));
  E_NULL_ERROR(obj);

  i = 0;
  aip = E_ASET_FIRST(asp);

                       /* Move to start */
  while((i < start) && aip)
  { aip = E_ASET_ITEM_NEXT(aip);
    i++;
  }
  
  while(aip)
  { if(E_EQUALS(E_ASET_ITEM_VAL(aip),obj)) 
    { E_FREE(obj);
      return(i);
    }
    aip = E_ASET_ITEM_NEXT(aip);
    i++;
  }
  E_FREE(obj);
                       /* Key not found */
  return(-1);

error:
  if(obj) E_FREE(obj);
  return(-2);  
}

/***********************************************************************/
/* Return the value associated with the key staring at position start. */
/* The first aset item has index 0.                                    */
/* Example :                                                           */    
/*  obj = e_aset_get_key_value(asp,e_name_new("test"),0);              */
/* asp : aset to scan.                                                 */
/* np : key name                                                       */
/* start : index of the first element to check                         */
/* Return : value associated with the key                              */
/*          0 : key not found or error                                 */
/* Link : asp  0                                                       */
/*        np  -1                                                       */
/*        Return : +1                                                  */
/***********************************************************************/
void *e_aset_get_value_by_key(e_aset *asp, e_name *np,int start)
{ int i; 
  e_aset_item *aip;

  E_NULL_ERROR(e_check_type(asp,E_TYPE_ASET,
                              "e_aset_get_value_by_key(asp,-,-)"));
  E_NULL_ERROR(e_check_type(np,E_TYPE_NAME,
                              "e_aset_get_value_by_key(-,np,-)"));

  i = 0;
  aip = E_ASET_FIRST(asp);

                       /* Move to start */
  while((i < start) && aip)
  { aip = E_ASET_ITEM_NEXT(aip);
    i++;
  }
  
  while(aip)
  { if(E_ASET_ITEM_NAME(aip) == np) 
    { E_FREE(np);
      if(E_ASET_ITEM_VAL(aip)) E_REF(E_ASET_ITEM_VAL(aip));
      return(E_ASET_ITEM_VAL(aip));
    }
    aip = E_ASET_ITEM_NEXT(aip);
    i++;
  }
  E_FREE(np);
                       /* Key not found */
  return(0);

error:
  if(np) E_FREE(np);
  return(0);  
}

/***********************************************************************/
/* Get an aset element value using its index position in the aset.     */
/* The first element has index 0.                                      */
/* Example :                                                           */    
/*  obj = e_aset_get_value_by_index(asp,0);                            */
/* asp : aset to scan.                                                 */
/* index : position of object, if negative, count from the end.        */
/*         0 : first element; -1 : last element.                       */
/*         if the index is invalid, it returns NULL but not error      */
/*         are generated.                                              */
/* Return : value associated with the key                              */
/*          0 : object not found                                       */
/*          0 : asp is not an aset.                                    */
/* Link : asp  0                                                       */
/*        Return : +1                                                  */
/***********************************************************************/
void *e_aset_get_value_by_index(e_aset *asp, int index)
{ int i,j; 
  e_aset_item *aip;

  E_NULL_ERROR(e_check_type(asp,E_TYPE_ASET,
                         "e_aset_get_value_by_index(asp,-)"));

  if(index < 0) i = E_ASET_CARDINAL(asp)+index;
  else i = index;

  if ((i < 0) || (i >= E_ASET_CARDINAL(asp)))
  {                          /* The aset member does not exist */
    return(0);
  }

  if (i > (E_ASET_CARDINAL(asp)/2))
  {                          /* second half of aset: start at the end */
    j = E_ASET_CARDINAL(asp)-1;
    aip = E_ASET_LAST(asp);
    while((j > i) && aip)
    { aip = E_ASET_ITEM_PREV(aip);
      j--;
    }
  }
  else
  {                         /* first half of aset */
    j = 0;
    aip = E_ASET_FIRST(asp);
    while((j < i) && aip)
    { aip = E_ASET_ITEM_NEXT(aip);
      j++;
    }
  }

  if(!aip)                 /* should never happen */
  { e_error(E_ERROR_INCONSISTENCY,E_ERROR_LEVEL_PANIC,
        "Aset inconsistency : fewer element than cardinal",(unsigned int)asp);
    return(0);       
  }

  if(E_ASET_ITEM_VAL(aip)) E_REF(E_ASET_ITEM_VAL(aip));
  return(E_ASET_ITEM_VAL(aip));

error:
  return(0);  
}

/******************************************************/
/* Return a pointer to the aset_item for nthe key np. */
/* To use with care.                                  */
/* asp : aset to scan.                                */
/* np : name for the key to search.                   */
/* Return : aset_item for key np.                     */
/*          0 : error or item not found               */
/* Link asp : 0                                       */
/*      np  : 0                                       */
/*      Return  : 0                                   */
/******************************************************/
e_aset_item *e_aset_get_item_by_key(e_aset *asp, e_name *np)
{ e_aset_item *aip;

  aip = E_ASET_FIRST(asp);

  while(aip)
  { if(E_ASET_ITEM_NAME(aip) == np) 
    { return(aip);
    }
    aip = E_ASET_ITEM_NEXT(aip);
  }
  return(0);
}

/*********************************************************/
/* Return a pointer to the aset_item for the object val. */
/* To use with care.                                     */
/* asp : aset to scan.                                   */
/* val : the object to search.                           */
/* Return : aset_item for val.                           */
/*          0 : error or item not found                  */
/* Link asp : 0                                          */
/*      val : 0                                          */
/*      Return  : 0                                      */
/*********************************************************/
e_aset_item *e_aset_get_item_by_value(e_aset *asp, void *val)
{ e_aset_item *aip;

  if(!val || !asp) return(0);

  aip = E_ASET_FIRST(asp);

  while(aip)
  { if(E_EQUALS(val,E_ASET_ITEM_NAME(aip))) 
    { return(aip);
    }
    aip = E_ASET_ITEM_NEXT(aip);
  }
  return(0);
}

/************************************************************************/
/* Split a string into a set of words. sep are the list of space        */
/* separators (as a string), while psep are the punctuation separators. */
/* By default sep = " \t" and psep = "".                                */
/*                                                                      */
/* Examples: sep= " ", psep = ",;"                                      */
/* "a b c"      => { "a" "b" "c" }                                      */
/* "a,b,c"      => { "a" "b" "c" }                                      */
/* "a,,c"       => { "a" "" "c" }                                       */
/* "a c"        => { "a" "c" }                                          */
/* "a , b , c"  => { "a" "b" "c" }                                      */
/* sp : string to split                                                 */
/* sep : space separators. Multiple space only count as one             */
/* psep : punctuation separator. Each one marks then end of a word      */
/* Return : the set of strings.                                         */
/*          0 : error                                                   */
/* Link : sp :     -1                                                   */
/*        Return : +1                                                   */
/************************************************************************/
e_aset *e_aset_split(e_string *sp,char *sep, char *psep)
{ e_tokeniser *tp = 0;
  e_aset *asp = 0;
  int state,r;

#define E_ASET_SPLIT_MODE_WORD 1
#define E_ASET_SPLIT_MODE_PUNC 2

  if(!sep) sep = " \t";
  if(!psep) psep = "";
  
  if(!(tp = e_tokeniser_new(0,e_charstream_new(0,
                                      E_CHARSTREAM_TYPE_STRING,sp,E_MODE_READ),
                            0," -~","",psep,sep)))
  { goto error;
  }

  E_NULL_ERROR(asp = e_aset_new(0)); 

  state = E_ASET_SPLIT_MODE_WORD;
  while((r = e_tokeniser_next(tp)) > 0)
  { 
    switch(tp->type)
    { case E_TOKENISER_WORD_TOKEN :
	                               /* Add word to set*/
           E_NULL_ERROR(asp = e_aset_append_item(asp,0,
                               e_string_new(0,E_TOKENISER_BUFF(tp),0)));  
           state = E_ASET_SPLIT_MODE_WORD; 
           break;

      case E_TOKENISER_OP_TOKEN :       /* read the = op */
           if(state == E_ASET_SPLIT_MODE_PUNC)
	   {                            /* Add empty string to set */
             E_NULL_ERROR(asp = e_aset_append_item(asp,0,
                               e_string_new(0,"",0))); 
           }

           state = E_ASET_SPLIT_MODE_PUNC;
           break;
    }
  }
  if(!r) goto error;

  return(asp);

error:
  if(tp)   E_FREE(tp);
  if(asp)  E_FREE(asp);
  return(0);
}

/**********************************************************************/ 
/* Add the property val (usualy a string but is can be any object)    */
/* to the property key.                                               */
/* asp is a list of properties, each identified by a key.             */
/* Each property list is an aset of objects.                          */
/* If the property is not in the list yet, it is added to asp. The    */
/* property list is the aset containing the single object obj. If val */
/* is null the property list is the empty aset.                       */
/* If the property key already exists its value is added to the       */
/* property list after checking that val is not already in the        */
/* property list.                                                     */
/* asp : a aset of aset : the aset of properties.                     */
/* key : a e_name : the property name (key in the aset)               */
/* mode : E_ASET_PROPERTY_ADD_SET : val is a set of properties to add */
/*        E_ASET_PROPERTY_MULTI : do not check if val is already in   */
/*               the property list (allow multiplicity values)        */
/* Return : the new property list.                                    */
/*          0 :error                                                  */
/* Link : asp  -1                                                     */
/*        key : -1                                                    */
/*        val : -1                                                    */
/*        Return : +1                                                 */
/**********************************************************************/ 
e_aset *e_aset_add_property(e_aset *asp, e_name *key, void *val,int mode)
{ e_aset *p,*aspv=0;
  e_aset_item *aip;

  if(!asp) 
  { e_error_null_pointer("e_aset_add_property(asp,-,-,-)");  
    goto error;
  }
  if(!key) 
  { e_error_null_pointer("e_aset_add_property(-,key,-,-)"); 
    goto error;
  }

             /* Make sure the aset is used only once     */
             /* So that we can modify it directly below! */
  if(E_OBJ_REF(asp) > 1) 
  { p = E_CLONE(asp,0); E_FREE(asp);
    asp = p;
  }


  if(!(aip = e_aset_get_item_by_key(asp,E_USE(key))))
  {                      /* Key not found : add  key = { val }  */
    return(e_aset_append_item(asp,key,
                              e_aset_append_item(e_aset_new(0),0,val)));
  }

              /* Is val already in the set ? */
  aspv = E_ASET_ITEM_VAL(aip);
  E_NULL_ERROR(e_check_type(aspv,E_TYPE_ASET,
      "e_aset_add_property : property item must be a set."));
  

  if(mode && E_ASET_PROPERTY_ADD_SET)
  { E_NULL_ERROR(e_check_type(val,E_TYPE_ASET,
      "e_aset_add_property(-,val,ADD_SET) : val must be a set."));
 
    if(mode & E_ASET_PROPERTY_MULTI)
    {                       /* Just append the list */
      E_ASET_ITEM_VAL(aip) = e_aset_append(aspv,val);
      return(asp);
    }
                            /* Add each element one by one. */
    E_ASET_ITEM_VAL(aip) = e_aset_add_aset(E_ASET_ITEM_VAL(aip),val,
                                             E_ASET_ADD_MODE_UNIQUE_VAL);
    return(asp);
  }


  if((mode & E_ASET_PROPERTY_MULTI) || 
     (e_aset_get_element_index(aspv,E_USE(val),0) < 0))
  {                      /* Property not yet in the set : add it */
    E_ASET_ITEM_VAL(aip) = e_aset_append_item(aspv,0,val);  
  }  
  return(asp);

error:
  if(aspv) E_FREE(aspv);  
  if(key) E_FREE(key);  
  if(val) E_FREE(val);  
  return(0);
}
