#include <easy.h>
#include <string.h>
#include <stdlib.h>
#include <stdio.h>
#include <sys/stat.h>
#include <unistd.h>

#ifndef fileno 
#ifdef LINUX
#define fileno(p)       ((p)->_fileno)
#endif
#ifdef SOLARIS
#define fileno(p)       ((p)->_file)
#endif
#endif

e_obj_descriptor e_obj_descriptor_charstream = {
  "charstream",       /* Name      */
  E_OBJ_MODE_NO_CLONE,/* def_mode  */
  malloc,             /* _alloc    */
  free,               /* _free     */
  realloc,            /* _realloc  */
  e_charstream_free,      /* free      */
  e_charstream_to_string, /* to_string */
  e_charstream_clone,     /* clone     */
  e_charstream_equals,    /* equals    */
};


/*****************************************************************************/
/* Create or initialise a char stream.                                       */
/* csp : static e_charstream to initialise. If NULL dynamically allocate one */
/* type : E_CHARSTREAM_TYPE_STRING, E_CHARSTREAM_TYPE_FILE or                */
/*        E_CHARSTREAM_TYPE_OTHER                                            */
/* obj : pointer to data :                                                   */
/*    File :  e_file   : file open for reading or writing.                   */
/*    String: e_string : initial value                                       */
/* mode : E_MODE_READ  : stream can be read                                  */
/*      : E_MODE_WRITE : stream can be written                               */
/* If csp is null, an e_charstream is allocated with malloc and the pointer  */
/* to the new object is returned. Use e_charstream_free to free it.          */
/* Return : e_charstream (csp or allocated one)                              */
/*          0 : error                                                        */
/* Link : obj     -1                                                         */
/*        Return  +1                                                         */
/*****************************************************************************/
e_charstream *e_charstream_new(e_charstream *csp,int type,void *obj, int mode)
{ char *fmode;

  if(!obj) return(0);

  if (!csp)
  { if (!(csp = e_obj_descriptor_charstream._alloc(sizeof(e_charstream)))) 
         return((e_charstream *)e_error_no_memory("e_charstream_new"));
    csp->h.refs = 1;
  }
  else csp->h.refs = -1;
  csp->h.mode = 0;

  csp->h.obj_desc_p = &e_obj_descriptor_charstream;

  csp->type = type;
  csp->mode = mode;
  csp->obj = obj;

  E_NULL_ERROR(csp->stack = e_stack_new(0,E_STACK_TYPE_INT,20));

  switch(type)
  { case E_CHARSTREAM_TYPE_OTHER : /* nothing is set */
              csp->flush = 0;
              csp->put = 0;
              csp->get = 0;
              csp->unget = 0;
              csp->seek = 0;
              csp->tell = 0;
           break;
    case E_CHARSTREAM_TYPE_STRING :
                                   /* Initialise the Methods */
              csp->flush = e_charstream_string_flush;
              csp->put   = e_charstream_string_put;
              csp->get   = e_charstream_string_get;
              csp->unget = e_charstream_unget;
              csp->seek  = e_charstream_string_seek;
              csp->tell  = e_charstream_tell;
	                           /* Rewind the stream */
              E_STRING_INDEX(E_CHARSTREAM_OBJ(csp)) = 0;
           break;
    case E_CHARSTREAM_TYPE_FILE :
                                   /* Initialise the Methods */
              csp->flush = e_charstream_file_flush;
              csp->put   = e_charstream_file_put;
              csp->get   = e_charstream_file_get;
              csp->unget = e_charstream_unget;
              csp->seek  = e_charstream_file_seek;
              csp->tell  = e_charstream_tell;
	                           /* Rewind the stream */
              rewind(E_FILE_FPT(E_CHARSTREAM_OBJ(csp))); 

	         /* Check the compatibility between the file and stream mode */
              fmode = E_FILE_MODE(E_CHARSTREAM_OBJ(csp));
              if(mode & E_MODE_READ)
	      { if((fmode[0] != 'r')&&(fmode[1] != '+'))
                { e_error(E_ERROR_FILE,E_ERROR_LEVEL_USER,
   "Incompatible stream mode READ for file (%s) %s",
                          E_FILE_FILENAME(E_CHARSTREAM_OBJ(csp)),fmode);
                  goto error;
		}
	      }
              if(mode & E_MODE_WRITE)
	      { if((fmode[0] != 'w')&&(fmode[0] != 'a')&&(fmode[1] != '+'))
                { e_error(E_ERROR_FILE,E_ERROR_LEVEL_USER,
   "Incompatible stream mode WRITE for file (%s) %s",
                          E_FILE_FILENAME(E_CHARSTREAM_OBJ(csp)),fmode);
                  goto error;
		}
	      }
  
           break;
    default : goto error;
           break;
  }
  return(csp);

error:

  if (obj) E_FREE(obj); 

  if(csp)
  { if (csp->stack) E_FREE(csp->stack); 
    if(csp->h.refs > 0)  e_obj_descriptor_charstream._free(csp);
  }
  return(0);
}

/*****************************************************/
/* Free the e_charstream                             */
/* Free the dynamicaly allocated data by calling the */
/* e_charstream.free function.                       */
/* Free the e_charstream if it is mallocated         */
/*****************************************************/
void e_charstream_free(void *csp)
{ int free = 1;

  if(csp) 
  { if(E_OBJ_REF(csp) == 0)
    { e_error(E_ERROR_INCONSISTENCY,E_ERROR_LEVEL_PANIC,
            "Trying to free a free charstream (%x) ",csp);
      e_error_flush(E_ERROR_LEVEL_DEBUG); 
      return;
    }

    if (E_OBJ_REF(csp) < 0)      /* static stack, free the buffer */
    { free = 0; 
      goto freeall;
    }

    if (E_OBJ(csp)->refs-- > 1) return;   /* decrement counter    */

freeall:
    if (E_CHARSTREAM_OBJ(csp))   
    { E_FREE(E_CHARSTREAM_OBJ(csp)); E_CHARSTREAM_OBJ(csp) = 0; 
    }
    if (E_CHARSTREAM_STACK(csp)) 
    { E_FREE(E_CHARSTREAM_STACK(csp)); E_CHARSTREAM_STACK(csp) = 0; 
    }

    if(free) e_obj_descriptor_charstream._free(csp);
  }
}

/*************************************************************/
/* Convert the file to a string                              */
/* The file is converted to the string                       */
/* Stream(OBJ,TYPE,MODE)                                     */
/* csp : a charstream, assumed to be of the correct type     */
/* Return : new string                                       */
/*          0 : error                                        */
/* LINK : csp : 0                                            */
/*        RETURN : +1                                        */
/*************************************************************/
void *e_charstream_to_string(void *csp,int tab,e_flag mode)
{ e_string *stp = 0;
  char *p;

  E_NULL_ERROR(stp = e_string_new(0,"Stream(",E_LINE_SIZE));

  E_NULL_ERROR(stp = e_string_append_string(stp,
               E_TO_STRING(E_CHARSTREAM_OBJ(csp),0,E_TO_STR_COMPACT)));

                	       /* add stream type */
  switch(E_CHARSTREAM_TYPE(csp))
  { case E_CHARSTREAM_TYPE_STRING: p = "\"String\""; break;
    case E_CHARSTREAM_TYPE_FILE: p = "\"File\""; break;
    case E_CHARSTREAM_TYPE_OTHER: p = "\"Other\""; break;
    default : p = "";
  }
  E_NULL_ERROR(stp = e_string_append_buffer(stp,p));
  
                	       /* add stream mode */
  switch(E_CHARSTREAM_TYPE(csp))
  { case E_MODE_READ: p = ",\"Read\")"; break;
    case E_MODE_WRITE: p = ",\"Write\")"; break;
    case (E_MODE_WRITE|E_MODE_READ): p = ",\"Read&Write\")"; break;
    default : p = "";
  }
  E_NULL_ERROR(stp = e_string_append_buffer(stp,p));

  if(!(mode & E_TO_STR_NO_TAB))
  { /* add tab to the string */
    stp = e_string_insert_chars(stp,0,tab,' ');
  }

  return (stp);

error:
  if (stp) E_FREE(stp);
  return(0);
}

/*****************************************************************/
/* Clone a charstream                                            */
/* If the char stream can't be cloned (default), return csp      */
/* and increase its refs counter                                 */
/* If cloned, the object (file or string)  is just reopened with */
/* the same attributes and the stream is rewinded.               */
/* csp : a file, assumed to be of the correct type               */
/* mode : cloning mode                                           */
/* Return : cloned e_charstream                                  */
/*          0 : Error                                            */
/* LINK : stp : 0                                                */
/*        RETURN : +1                                            */
/*****************************************************************/
void *e_charstream_clone(void *csp,e_flag mode)
{ e_charstream *newcsp = 0;
  e_obj *newobj = 0;
  e_flag obj_mode;

                 /* Can't clone : return the same object */
  if(E_OBJ_NO_CLONE(csp)) return(E_USE(csp));

                 /* save the object mode */
  obj_mode = E_OBJ_MODE(E_CHARSTREAM_OBJ(csp));

                 /* make the stream object clonable */
                 /* The object is always cloned     */
  E_OBJ_MODE(E_CHARSTREAM_OBJ(csp)) &= ~E_OBJ_MODE_NO_CLONE;

                 /* clone the object an create a new charstream */
  if(!(newobj = E_CLONE(E_CHARSTREAM_OBJ(csp),mode)))
  { E_OBJ_MODE(E_CHARSTREAM_OBJ(csp)) = obj_mode;
    goto error;
  }

                 /* restore the object mode */
  E_OBJ_MODE(E_CHARSTREAM_OBJ(csp)) = obj_mode;

                 /* create a new stream from the cloned object */
  if(!(newcsp = e_charstream_new(0,E_CHARSTREAM_TYPE(csp),
                                  newobj,E_CHARSTREAM_MODE(csp)))) 
      goto error2;
 
                 /* Ohter stream : copy the methods too */ 
  if(E_CHARSTREAM_TYPE(csp) == E_CHARSTREAM_TYPE_OTHER)
  { newcsp->flush = ((e_charstream *)csp)->flush;
    newcsp->put = ((e_charstream *)csp)->put;
    newcsp->get = ((e_charstream *)csp)->get;
    newcsp->unget = ((e_charstream *)csp)->unget;
    newcsp->seek = ((e_charstream *)csp)->seek;
    newcsp->tell = ((e_charstream *)csp)->tell;
  }

  return(newcsp);

error:
  if(newobj) E_FREE(newobj); 
error2:
  if(newcsp) E_FREE(newcsp); 
  return(0);
}

/*******************************************************************/
/* Test if the 2 charstream are equal:                             */
/* The two charstreams are equal if they are identical (p1 == p2), */
/* or if they have the same object (file, string, ...)             */
/* and their mode, types and index are indentical.                 */
/* Return : 1 : p1 and p2 are equal                                */
/*          0 : p1 and p2 are different                            */
/* Link : p1 0                                                     */
/*        p2 0                                                     */
/*******************************************************************/
int e_charstream_equals(void *p1,void *p2)
{ 

  if (p1 == p2)   return(1);
  if (!p1 || !p2) return(0);

            /* different object types => differ */
  if(E_OBJ_DESCR(p1) != E_OBJ_DESCR(p2)) return(0);

            /* different index => differ */
  if(E_CHARSTREAM_INDEX(p1) != E_CHARSTREAM_INDEX(p2)) return(0);

            /* different types => differ */
  if(E_CHARSTREAM_TYPE(p1) != E_CHARSTREAM_TYPE(p2)) return(0);

            /* different modes => differ */
  if(E_CHARSTREAM_MODE(p1) != E_CHARSTREAM_MODE(p2)) return(0);

                 /* Other stream : compare the methods too */ 
  if((E_CHARSTREAM_TYPE(p1) == E_CHARSTREAM_TYPE_OTHER) &&
    ((((e_charstream *)p1)->flush != ((e_charstream *)p2)->flush)||
     (((e_charstream *)p1)->put != ((e_charstream *)p2)->put)||
     (((e_charstream *)p1)->get != ((e_charstream *)p2)->get)||
     (((e_charstream *)p1)->unget != ((e_charstream *)p2)->unget)||
     (((e_charstream *)p1)->seek != ((e_charstream *)p2)->seek)||
     (((e_charstream *)p1)->tell != ((e_charstream *)p2)->tell))) 
             return(0);
  
            /* then compare the objects */
  return(E_EQUALS(E_CHARSTREAM_OBJ(p1),E_CHARSTREAM_OBJ(p2)));

}

/******************************************************/
/* Unget one char on the stream :                     */
/* The character is pushed onto a stack.              */ 
/* Is is not possible to unget characters beyond the  */
/* start of the file.                                 */
/* Ungeting EOF has no effect, but is allowed         */
/* csp : char stream to unget onto                    */
/* c   : the character to unget, it does not need to  */
/*       the last character read.                     */
/* Return : 1 : OK                                    */
/*          0 : error                                 */
/* Link : csp  0                                      */
/******************************************************/
int e_charstream_unget(e_charstream *csp,int c)
{ E_NULL_ERROR(e_check_type(csp,E_TYPE_CHARSTREAM,"e_charstream_unget"));

  if(c == EOF) return(1);

  if(E_CHARSTREAM_INDEX(csp) <= 0) return(0);

  E_NULL_ERROR(e_stack_push(E_CHARSTREAM_STACK(csp),&c));

  E_CHARSTREAM_INDEX(csp)--;
  return(1);

error:
  return(0);
}

/******************************************************/
/* Return the index position in the stream.           */
/* csp : e_charstream to look at                      */
/* Return : stream index position                     */
/*          -1 : error                                */
/* Link : csp  0                                      */
/******************************************************/
long e_charstream_tell(e_charstream *csp)
{ 
  E_NULL_ERROR(e_check_type(csp,E_TYPE_CHARSTREAM,"e_charstream_tell"));

  return(E_CHARSTREAM_INDEX(csp));

error:
  return(-1);
}

/********************** STRING CHAR STREAM ******************/

/*****************************************************/
/* flush the char stream : nothing to do for strings */ 
/* csp : char stream to flush                        */
/* Return : 1 : OK                                   */
/*          0 : error                                */
/* Link : csp  0                                     */
/*****************************************************/
int e_charstream_string_flush(e_charstream *csp)
{ return(1);
}

/******************************************************/
/* Put one char on the stream :                       */
/* csp : char stream to read                          */
/* c : char to write                                  */
/* Return : character read                            */
/*          EOF : end of file reached.                */
/*          0 : error                                 */
/* Link : csp  0                                      */
/******************************************************/
int e_charstream_string_put(e_charstream *csp,int c)
{ 
  E_NULL_ERROR(e_check_type(csp,E_TYPE_CHARSTREAM,"e_charstream_put"));

  if(!(E_CHARSTREAM_MODE(csp) & E_MODE_WRITE))
  { e_error(E_ERROR_PROTECTION,E_ERROR_LEVEL_USER,
  "Charstream write error: string charstream \"%s\" is open for reading only",
             E_STRING_S(E_CHARSTREAM_OBJ(csp)));
    goto error;
  }

  E_NULL_ERROR(e_string_iput((e_string *)E_CHARSTREAM_OBJ(csp),c));

  E_CHARSTREAM_INDEX(csp)++;

  return(c);

error:
  return(0);
}

/******************************************************/
/* Get one char on the stream :                       */
/* Test if any character has been pushed on the stack */
/* csp : char stream to read                          */
/* Return : character read                            */
/*          EOF : end of file reached.                */
/*          0 : error                                 */
/* Link : csp  0                                      */
/******************************************************/
int e_charstream_string_get(e_charstream *csp)
{ int c;

  E_NULL_ERROR(e_check_type(csp,E_TYPE_CHARSTREAM,"e_charstream_get"));

  if(!(E_CHARSTREAM_MODE(csp) & E_MODE_READ))
  { e_error(E_ERROR_PROTECTION,E_ERROR_LEVEL_USER,
  "Charstream read error: string charstream \"%s\" is open for writing only",
             E_STRING_S(E_CHARSTREAM_OBJ(csp)));
    goto error;
  }

  if(E_STACK_IS_EMPTY(E_CHARSTREAM_STACK(csp)))
  { c = e_string_iget((e_string *)E_CHARSTREAM_OBJ(csp));
  }
  else
  { E_NULL_ERROR(e_stack_pop(E_CHARSTREAM_STACK(csp),&c));
  }
  E_CHARSTREAM_INDEX(csp)++;

  return(c);

error:
  return(0);
}

/***********************************************************************/
/* Change the index position in the stream.                            */
/* Moving beyond the beginning or end of stream generates an error.    */
/* csp : e_charstream to look at                                       */
/* pos : new position (can be negative with E_MODE_MOVE_REL)           */
/* mode : seek mode:                                                   */
/*        E_MODE_MOVE_ABS : from begining. 0 is 1st char, len=EOS      */
/*        E_MODE_MOVE_CUR : from current position, pos can be < 0      */
/*        E_MODE_MOVE_END : from end. 0=EOS, 1=last char, len=1st char */
/* Return : 1 : OK                                                     */
/*          0 : error                                                  */
/* Link : csp  0                                                       */
/***********************************************************************/
int e_charstream_string_seek(e_charstream *csp,long pos,int mode)
{ int newpos=0,len;

  E_NULL_ERROR(e_check_type(csp,E_TYPE_CHARSTREAM,"e_charstream_string_seek"));


  len = E_STRING_LEN(E_CHARSTREAM_OBJ(csp));

  switch (mode)
  { case E_MODE_MOVE_ABS : 
             newpos = pos;             
             break;
    case E_MODE_MOVE_REL : 
             newpos = E_CHARSTREAM_INDEX(csp)+pos;             
             break; 
    case E_MODE_MOVE_END : 
             newpos = len-pos;             
             break; 
  }

  if((newpos < 0) || (newpos > len))  
  { e_error(E_ERROR_OUT_OF_BOUND,E_ERROR_LEVEL_USER,
            "Invalid seek (%d) value for string stream %s",newpos,
            E_STRING_S(E_CHARSTREAM_OBJ(csp)));
    return(0);
  }

  E_CHARSTREAM_INDEX(csp) = newpos;
  E_STRING_INDEX(E_CHARSTREAM_OBJ(csp)) = newpos;

  return(1); 

error: 
  return(0); 
}

/********************** FILE CHAR STREAM ******************/

/*****************************************************/
/* flush the char stream :                           */ 
/* csp : char stream to flush                        */
/* Return : 1 : OK                                   */
/*          0 : error                                */
/* Link : csp  0                                     */
/*****************************************************/
int e_charstream_file_flush(e_charstream *csp)
{ 
  E_NULL_ERROR(e_check_type(csp,E_TYPE_CHARSTREAM,"e_charstream_file_flush"));

  fflush(E_FILE_FPT(E_CHARSTREAM_OBJ(csp)));
  return(1);

error:
  return(0);
}

/******************************************************/
/* Put one char on the stream :                       */
/* csp : char stream to read                          */
/* c : char to write                                  */
/* Return : character read                            */
/*          EOF : end of file reached.                */
/*          0 : error                                 */
/* Link : csp  0                                      */
/******************************************************/
int e_charstream_file_put(e_charstream *csp,int c)
{ 
  E_NULL_ERROR(e_check_type(csp,E_TYPE_CHARSTREAM,"e_charstream_put"));

  if(!(E_CHARSTREAM_MODE(csp) & E_MODE_WRITE))
  { e_error(E_ERROR_PROTECTION,E_ERROR_LEVEL_USER,
    "Charstream write error: file charstream \"%s\" is open for reading only",
             E_FILE_FILENAME(E_CHARSTREAM_OBJ(csp)));
    goto error;
  }

  if(EOF == fputc(c,E_FILE_FPT(E_CHARSTREAM_OBJ(csp))))
  { e_error(E_ERROR_FILE,E_ERROR_LEVEL_USER,
            "Charstream write error: error writing into the file \"%s\"",
             E_FILE_FILENAME(E_CHARSTREAM_OBJ(csp)));
    goto error;
  }

  E_CHARSTREAM_INDEX(csp)++;

  return(c);

error:
  return(0);
}

/******************************************************/
/* Get one char on the stream :                       */
/* Test if any character has been pushed on the stack */
/* csp : char stream to read                          */
/* Return : character read                            */
/*          EOF : end of file reached.                */
/*          0 : error                                 */
/* Link : csp  0                                      */
/******************************************************/
int e_charstream_file_get(e_charstream *csp)
{ int c;

  E_NULL_ERROR(e_check_type(csp,E_TYPE_CHARSTREAM,"e_charstream_get"));

  if(!(E_CHARSTREAM_MODE(csp) & E_MODE_READ))
  { e_error(E_ERROR_PROTECTION,E_ERROR_LEVEL_USER,
    "Charstream read error: file charstream \"%s\" is open for writing only",
             E_FILE_FILENAME(E_CHARSTREAM_OBJ(csp)));
    goto error;
  }

  if(E_STACK_IS_EMPTY(E_CHARSTREAM_STACK(csp)))
  { c = fgetc(E_FILE_FPT(E_CHARSTREAM_OBJ(csp)));
    if (ferror(E_FILE_FPT(E_CHARSTREAM_OBJ(csp))))
    { e_error(E_ERROR_FILE,E_ERROR_LEVEL_USER,
            "Charstream read error: error reading the file \"%s\"",
             E_FILE_FILENAME(E_CHARSTREAM_OBJ(csp)));
      goto error;
    }
  }
  else
  { E_NULL_ERROR(e_stack_pop(E_CHARSTREAM_STACK(csp),&c));
  }
  E_CHARSTREAM_INDEX(csp)++;

  return(c);

error:
  return(0);
}

/***********************************************************************/
/* Change the index position in the stream.                            */
/* Moving beyond the beginning or end of stream generates              */
/* an error.                                                           */
/* csp : e_charstream to look at                                       */
/* pos : new position (can be negative with E_MODE_REL)                */
/* mode : seek mode:                                                   */
/*        E_MODE_MOVE_ABS : from begining. 0 is 1st char, len=EOS      */
/*        E_MODE_MOVE_CUR : from current position, pos can be < 0      */
/*        E_MODE_MOVE_END : from end. 0=EOS, 1=last char, len=1st char */
/* Return : 1 : OK                                                     */
/*          0 : error                                                  */
/* Link : csp  0                                                       */
/***********************************************************************/
int e_charstream_file_seek(e_charstream *csp,long pos,int mode)
{ int newpos=0,len,fd,seek_mode=0;
  struct stat stp;
  char *s=0;

  E_NULL_ERROR(e_check_type(csp,E_TYPE_CHARSTREAM,"e_charstream_string_seek"));

  if(((E_FILE_FPT(E_CHARSTREAM_OBJ(csp)) == stdout) && (s == "stdout" )) ||
     ((E_FILE_FPT(E_CHARSTREAM_OBJ(csp)) == stdout) && (s == "stdout")) ||
     ((E_FILE_FPT(E_CHARSTREAM_OBJ(csp)) == stderr) && (s == "stderr")))
  { e_error(E_ERROR_FILE,E_ERROR_LEVEL_USER,
            "Charstream seek error: can't seek %s!",s);
    goto error; 
  }

  /* We get the file size using fstat */
  fd = fileno(E_FILE_FPT(E_CHARSTREAM_OBJ(csp)));
  if(fstat(fd,&stp))
  { e_error(E_ERROR_FILE,E_ERROR_LEVEL_USER,
            "Charstream seek error: file \"%s\" : fstat error : %s",
              E_FILE_FILENAME(E_CHARSTREAM_OBJ(csp)),strerror(errno));
    goto error; 
  }

  len = stp.st_size;

  switch (mode)
  { case E_MODE_MOVE_ABS : 
             newpos = pos; 
             seek_mode = SEEK_SET;            
             break;
    case E_MODE_MOVE_REL : 
             newpos = E_CHARSTREAM_INDEX(csp)+pos; 
             seek_mode = SEEK_CUR;           
             break; 
    case E_MODE_MOVE_END : 
             newpos = len-pos;             
             seek_mode = SEEK_END;           
             break; 
  }

  if((newpos < 0) || (newpos > len))  
  { e_error(E_ERROR_OUT_OF_BOUND,E_ERROR_LEVEL_USER,
            "Invalid seek (%d) value for file stream %s",newpos,
            E_STRING_S(E_CHARSTREAM_OBJ(csp)));
    return(0);
  }

  if(fseek(E_FILE_FPT(E_CHARSTREAM_OBJ(csp)),newpos,seek_mode))
  { e_error(E_ERROR_FILE,E_ERROR_LEVEL_USER,
            "Charstream seek error: file \"%s\" : fseek error : %s",
             E_FILE_FILENAME(E_CHARSTREAM_OBJ(csp)),strerror(errno));
    goto error;
  }

  E_CHARSTREAM_INDEX(csp) = newpos;

  return(1); 

error: 
  return(0); 
}
