#include <e_error.h>
#include <stdarg.h>
#include <stdio.h>
#include <easy.h>

      /* Statically allocated array for error messages */
      /* We can't rely on dynamic allocation here!     */
e_error_log e_error_list[E_LOG_SIZE]; 
int e_error_index; 
e_flag e_error_cur;        /* Current errors */

/***************************/
/* Initialisation function */
/***************************/
int e_error_init(void *p)
{ e_error_index = 0;
  return(1);
}

/*******************************************************/
/* Write formated error messages in the error buffer   */
/* level : error level (cfr definitions in e_error.h)  */
/*******************************************************/
void e_error(int type,int level,const char *fmt, ...)
{ va_list args;

  e_error_list[e_error_index].log_level = level;
  e_error_list[e_error_index].type      = type;
  e_error_cur |=  type;

  va_start(args, fmt);
  vsprintf(&(e_error_list[e_error_index].text[0]),fmt, args); 
  va_end(args);

  if(e_error_index < (E_LOG_SIZE-1)) e_error_index++;
}

/**************************************************/
/* Display error messages on stderr.              */
/* Only display error messages with log levels    */
/* less or equal than level                       */
/**************************************************/
void e_error_display(int level)
{ int i;

  for(i=0; i < e_error_index; i++)
  { if(e_error_list[i].log_level <= level) 
    { fprintf(stderr,"%s\n",e_error_list[i].text);
    }
  }
}

/**************************************************/
/* Return the error message into an e_string.     */
/* Only return error messages with log levels     */
/* less or equal than level                       */
/* Return : the error e_string                    */
/*          0 : error                             */
/* Link : Return +1                               */
/**************************************************/
void *e_error_2_string(int level)
{ int i;
  e_string *sp;

  sp = e_string_new(0,0,E_LONG_BUFF_SIZE);
  for(i=0; i < e_error_index; i++)
  { if(e_error_list[i].log_level <= level) 
    { sp = e_string_append_buffer(sp,e_error_list[i].text);
      sp = e_string_append_buffer(sp,"\n"); 
    }
  }
  return(sp);
}

/**********************/
/* Reset error buffer */
/**********************/
void e_reset()
{ e_error_index = 0;
  e_error_cur = 0;
}

/**************************************************/
/* Display error messages on stderr.              */
/* Only display error messages with log levels    */
/* less or equal than level                       */
/* Reset the error buffer                         */
/**************************************************/
void e_error_flush(int level)
{ 
  e_error_display(level);
  e_reset();
}

/*******************************************************/
/* Write formated synatx error messages at level USER  */
/* in the error buffer                                 */
/*******************************************************/
void e_error_syntax(const char *fmt, ...)
{ va_list args;

  e_error_list[e_error_index].log_level = E_ERROR_LEVEL_USER;
  e_error_list[e_error_index].type      = E_ERROR_SYNTAX;
  e_error_cur |=  E_ERROR_SYNTAX;

  va_start(args, fmt);
  vsprintf(&(e_error_list[e_error_index].text[0]),fmt, args); 
  va_end(args);

  if(e_error_index < (E_LOG_SIZE-1)) e_error_index++;
}

/***************************************************/
/* Log a null pointer error                        */
/* fct : the name of the function where it happens */
/* Return : 0                                      */
/* Usage :                                         */
/*   if(!p) return(e_error_null_pointer("My Fct"));*/
/***************************************************/
int e_error_null_pointer(char *fct)
{ e_error(E_ERROR_NULL_POINTER,E_ERROR_LEVEL_ADMIN,
         "%s : null pointer error",fct);
  return(0);
}

/***************************************************/
/* Log a memory error                              */
/* fct : the name of the function where it happens */
/* Return : 0                                      */
/* Usage :                                         */
/*   if(!p) return(e_error_memory("My Fct"));      */
/***************************************************/
int e_error_no_memory(char *fct)
{ e_error(E_ERROR_MEMORY,E_ERROR_LEVEL_PANIC,
         "%s : out of memory error!",fct);
  return(0);
}

/***************************************************/
/* Test for write protection error                 */
/* Test for null pointer as well                   */
/* p : object pointer to test                      */
/* fct : the name of the function where it happens */
/* Return : 0 : write or null pointer error        */
/*          1 : OK                                 */
/* Usage :                                         */
/*  E_NULL_ERROR(e_error_is_writable(p,"MyFct"));  */
/***************************************************/
int e_error_is_writable(void *p,char *fct)
{ 
                      /* is p NULL */
  if(!p) return(e_error_null_pointer(fct));
  
                      /* is p WRITE PROTECTED */
  if(E_OBJ_READ_ONLY(p))
  { e_error(E_ERROR_PROTECTION,E_ERROR_LEVEL_ADMIN,
         "%s : write protection error for %s (%x)",fct,E_OBJ_NAME(p),p);
    return(0); 
  }

                      /* can write into p */
  return(1);
}

/***************************************************/
/* Test for read protection error                  */
/* Test for null pointer as well                   */
/* p : object pointer to test                      */
/* fct : the name of the function where it happens */
/* Return : 0 : write or null pointer error        */
/*          1 : OK                                 */
/* Usage :                                         */
/*   if(!p) return(e_error_null_pointer("My Fct"); */
/***************************************************/
int e_error_is_readable(void *p,char *fct)
{ 
                      /* is p NULL */
  if(!p) return(e_error_null_pointer(fct));
  
                      /* is p READ PROTECTED */
  if(E_OBJ_NO_READ(p))
  { e_error(E_ERROR_PROTECTION,E_ERROR_LEVEL_ADMIN,
         "%s : read protection error for %s (%x)",fct,E_OBJ_NAME(p),p);
    return(0); 
  }

                      /* can write into p */
  return(1);
}

/****************************************************/
/* Check that p is non null and of the correct type */
/* Log an appropriate error.                        */
/* p : object to test                               */
/* odp : object type (pointer object decriptor)     */
/* fct : the name of the function where it happens  */
/* Return : 0 : error                               */
/*          1 : OK                                  */
/* Usage :                                          */
/*   E_ON_ERROR(e_check_type(p,E_STRING,"MyFct"));  */
/****************************************************/
int e_check_type(void *p,e_obj_descriptor *type,char *fct)
{ 
  if(!p)
  { e_error(E_ERROR_NULL_POINTER,E_ERROR_LEVEL_ADMIN,
           "%s : null pointer error",fct);
    return(0);
  }
  else if(E_OBJ_DESCR(p) != type)
  { e_error(E_ERROR_INVALID_OBJECT,E_ERROR_LEVEL_ADMIN,
           "%s : invalid object, %s expected",fct,type->Name);
    return(0);
  }
  return(1);
}
