#include <easy.h>
#include <stdio.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/wait.h>

#ifndef fdopen
#ifdef LINUX
extern FILE *fdopen __P ((int __fd, __const char *__modes));
#endif
#ifdef SOLARIS
extern FILE *fdopen(int, const char *);
#endif
#endif


/****************************************************************/
/* Execute the sh command com and return the output as a string */
/* If an error occurs, *err is set to 1 and the string returned */
/* is the error messages corresponding to errno                 */
/* Err is otherwise is set to 0.                                */
/* com : command to execute.                                    */
/* *err pointer to error flag. Is set to 1 wheen popen fails    */
/*    and to 0 otherwise.                                       */ 
/* if err is NULL, the function returns NULL when popen fails   */
/* The function does not log any error                          */
/* Return output string or error message                        */
/*        0 : error                                             */
/* Link : Return +1                                             */
/****************************************************************/
e_string *e_exec_sh_command(char *com,int *err)
{ FILE *pfpt = 0;
  e_string *output;
  char Buff[E_LONG_BUFF_SIZE+1];
  int s;
                    /* Execute the command */
  if(!(pfpt = (FILE *)popen(com,"r")))
  { if(err)         /* failed              */
    { *err = 1;
      if(errno) return(e_string_new(0,strerror(errno),0));

      sprintf(Buff,"Unable to execute %s\n",com);
      return(e_string_new(0,Buff,0));
    }
                    /* Return 0 to indicate the error */
    return(0);
  } 

  E_NULL_ERROR(output = e_string_new(0,"",E_LONG_BUFF_SIZE));

                     /* Read the output in blocks */
  while((s = fread(Buff,1,E_LONG_BUFF_SIZE,pfpt)))
  { Buff[s] = 0;
                     /* append blocks to the output string */
    E_NULL_ERROR(output = e_string_append_buffer(output,Buff));
  }

  pclose(pfpt);

  if(err) *err = 0; 
  return(output);

error:
  if (pfpt) pclose(pfpt);
  return(0);
}

/***************************************************************/
/* Execute the command argv to pipe write into it              */
/* Write a file descriptor open for writing into fptp.         */
/* The forked process reads what will be written into the file */
/* pointer.                                                    */
/* Finish by closing the file and wait for the PID returned    */
/* Sample code:                                                */
/*                                                             */
/*   e_exec_pipe_fwrite(argv,&fpt);                            */
/*   while (c=str[i++]) { fputc(c,fpt); }                      */
/*   fclose(fpt);                                              */
/*   status = e_exec_wait_for(pid,&sig,&stat);                 */
/*                                                             */
/* argv: argument list                                         */
/* fptp: pointer to file pointer                               */
/* Return : PID for child process                              */
/*          0 : error                                          */
/***************************************************************/
int e_exec_pipe_fwrite(char **argv, FILE **fptp)
{ int pfd[2];
  int pid;

  if (pipe(pfd)== -1) return(0);
  switch (pid = fork())
  { case -1 : return(0);
    case  0 : /* child */
        /* make pipe_in  stdin for exec! */
        if( close(0) == -1) return(0);
        if (dup(pfd[0]) != 0) exit(0);
        if (close(pfd[0]) == -1 || close(pfd[1]) == -1) return(0);

        execvp(argv[0],argv);
        return(0);
  }

  /* Parent : out pipe : => stdout.: close in pipe */
  if (close(pfd[0]) != 0) return(0);

  /* Get FILE* for output pipe */
  *fptp = (FILE *) fdopen(pfd[1],"w");

  return(pid);

}

/***************************************************************/
/* Execute the command argv and write data into it             */
/* The forked process reads data from stdin.                   */
/*                                                             */
/* argv: argument list                                         */
/* data: input data                                            */
/* Return : 1 : OK                                             */
/*          0 : error                                          */
/* Link : data : -1                                            */
/***************************************************************/
int e_exec_pipe_write(char **argv, e_string *data)
{ int pid,i,c;
  FILE *fpt;

  if(!argv || !data) goto error;
  pid = e_exec_pipe_fwrite(argv,&fpt);                            
  i = 0;
  
  while ((c=data->s[i])&&(i < data->len)) 
  { fputc(c,fpt); 
    i++;
  }
  fclose(fpt);
  e_exec_wait_for(pid,0,0);
  if(data) E_FREE(data);
  return(1);

error:
  if(data) E_FREE(data);
  return(0);
}

/***************************************************************/
/* Execute the command argv to pipe to read from it.           */
/* Write a file descriptor open for reading into fptp.         */
/* The forked process write into the file.                     */
/* After the call, read from fptp until EOF is reached.        */
/* Sample code:                                                */
/*                                                             */
/*   e_exec_pipe_fread(argv,&fpt);                             */
/*   while ((c = getc(*fptp)!= EOF)) { str[i++] = c;}          */
/*    str[i] = 0; fclose(*fptp);                               */
/*   status = OS_waitfor(pid,&sig,&stat);                      */
/*                                                             */
/* argv: argument list                                         */
/* fptp: pointer to file pointer                               */
/* Return : PID for child process                              */
/*          0 : error                                          */
/***************************************************************/
int e_exec_pipe_fread(char **argv, FILE **fptp)
{ int pfd[2];
  int pid;

  if (pipe(pfd)== -1) return(0);
  switch (pid = fork())
  { case -1 : return(0);
    case  0 : /* child */
        /* make pipe_out  stdout for exec! */
        if( close(1) == -1) return(0);
        if (dup(pfd[1]) != 1) exit(0);
        if (close(pfd[0]) == -1 || close(pfd[1]) == -1) return(0);

        execvp(argv[0],argv);
        return(0);
  }

  /* Parent : in pipe : => stdout.: close out pipe */
  if (close(pfd[1]) != 0) return(0);

  /* Get FILE* for input pipe */
  *fptp = (FILE *) fdopen(pfd[0],"r");

  return(pid);
}

/*********************************************************************/
/* Execute the command argv and return the stdin output as a string. */
/* argv: argument list                                               */
/* Return : Command output                                           */
/*          0 : error                                                */
/*********************************************************************/
e_string *e_exec_pipe_read(char **argv)
{ e_string *esp=0;
  int pid,n;
  char Buff[E_BUFF_SIZE+1];
  FILE *fpt;

  if(!argv) goto error;
  pid = e_exec_pipe_fread(argv,&fpt);                            
  
  E_NULL_ERROR(esp = e_string_new(0,"",E_BUFF_SIZE));
  while ((n = fread(Buff,1,E_BUFF_SIZE,fpt))) 
  { Buff[n] = 0; 
    E_NULL_ERROR(esp = e_string_append_buffer(esp,Buff));
  }
  fclose(fpt);
  e_exec_wait_for(pid,0,0);

  return(esp);

error:
  return(0);
}

/************************************************************/
/* Wait for process pid and return status code.             */
/* pid : the process to wait for.                           */
/* Set stat to exit status. If the process was  killed, the */
/* variable sig is set to the signal number received and    */
/* stat is set to 256.                                      */
/* Return : status as returned by wait.                     */
/************************************************************/
int e_exec_wait_for(int pid,int *sig, int *stat)
{ int wpid,status;

  while((wpid = wait(&status)) != pid && wpid != -1);

  if(sig) *sig = 0;
  /* if 1st bit of signal is NULL, 2nd bit of signal is error no */
  /* else 1st bit of signal is signal that stopped the process.  */
  if ((status & 0x7f) == 0x7f)
  { if(sig) *sig = status >> 8;
    if(stat) *stat = 0x7f;
  }
  else if (status & 0xff)
  { if(sig) *sig = status & 0x7f;
    if(stat) *stat = 256;
  }
  else
  { if(stat) *stat = status >> 8;
  }

  return(status);
}
