#include <easy.h>
#include <stdlib.h>

e_obj_descriptor e_obj_descriptor_number = {
  "number",             /* Name      */
  E_OBJ_MODE_NO_CLONE,  /* def_mode  */
  malloc,               /* _alloc    */
  free,                 /* _free     */
  realloc,              /* _realloc  */
  e_number_free,        /* free      */
  e_number_to_string,   /* to_string */
  e_number_clone,       /* clone     */
  e_number_equals,      /* equals    */
};


/************************************************************/
/* Number :                                                 */
/* Dynamic encapsulator for numerical values                */
/************************************************************/

/***************************************************************************/
/* Create or initialise a new e_number                                     */
/* np : e_number to initialise (static e_number)                           */
/*      if NULL allocate it                                                */
/*      np is assumed to be of the correct type                            */
/* type : type of number : E_NUM_TYPE_CHAR, E_NUM_TYPE_INT,                */
/*                         E_NUM_TYPE_REAL, E_NUM_TYPE_COMP                */
/* init : initial value as an integer.                                     */
/* Return e_number                                                         */
/*        0 : error                                                        */
/* Link : np -1                                                            */
/*        Return +1                                                        */
/***************************************************************************/
e_number *e_number_new(e_number *np,int type,int init)
{ 
                 /* Static object or allocate it ? */
  if (!np)
  { if (!(np = e_obj_descriptor_number._alloc(sizeof(e_number)))) 
      return((e_number *)e_error_no_memory("e_number_new"));
    np->h.refs = 1;
  }
  else np->h.refs = -1;
  np->h.mode = 0;

  np->h.obj_desc_p = &e_obj_descriptor_number;

                 /* Check for valid number type */
  if ((type < E_NUM_TYPE_CHAR)||(type > E_NUM_TYPE_COMP))
  { e_error(E_ERROR_INVALID_OBJECT,E_ERROR_LEVEL_USER,
            "e_number_new : Invalid numerical type (%d)",type);
    goto error;
  }
  np->type = type;

  switch(type)
  { case E_NUM_TYPE_CHAR : E_NUMBER_VAL_CHAR(np) = (char) init; break;
    case E_NUM_TYPE_INT :  E_NUMBER_VAL_INT(np) = (int) init; break;
    case E_NUM_TYPE_REAL : E_NUMBER_VAL_REAL(np) = (double) init; break;
    case E_NUM_TYPE_COMP : E_NUMBER_VAL_COMP(np).re = (double) init;
                           E_NUMBER_VAL_COMP(np).im = 0.0; 
                           break;
  }

  return(np);

error:
  if(np && (np->h.refs > 0)) e_obj_descriptor_number._free(np);

  return(0); 
}

/***************************************************************************/
/* Free the e_number                                                       */
/* If the refs counter is -1 : static number : leave as is                 */
/* If the refs counter is 1 : free the allocated memory for the e_number   */
/* Otherwise decrement the counter by 1.                                   */
/* REF : stp : -1                                                          */
/***************************************************************************/
void e_number_free(void *np)
{

  if(np)
  { if(E_OBJ_REF(np) == 0)
    { e_error(E_ERROR_INCONSISTENCY,E_ERROR_LEVEL_PANIC,
            "Trying to free a free number (%x) ",np);
      goto error;
    }
                                /* static number */
    if (E_OBJ_REF(np) < 0) return;

                                /* decrement the refs counter   */
    if (E_OBJ(np)->refs-- > 1) return;

                                /* Last reference,   */ 
                                /* free the e_number */
    e_obj_descriptor_number._free(np);
  }
  return;

error:
  e_error_flush(E_ERROR_LEVEL_DEBUG); 
  
}

/*************************************************************/
/* Convert the number to a string                            */
/* np : an e_number, assumed to be of the correct type       */
/* Return : new string                                       */
/*          0 : error                                        */
/* LINK : np : 0                                             */
/*        RETURN : +1                                        */
/*************************************************************/
void *e_number_to_string(void *np,int tab,e_flag mode)
{ e_string *stp = 0;
  char buff[E_LINE_SIZE];

  switch(E_NUMBER_TYPE(np))
  { case E_NUM_TYPE_CHAR : sprintf(buff,"%c",E_NUMBER_VAL_CHAR(np)); break;
    case E_NUM_TYPE_INT :  sprintf(buff,"%d",E_NUMBER_VAL_INT(np)); break;
    case E_NUM_TYPE_REAL : sprintf(buff,"%g",E_NUMBER_VAL_REAL(np)); break;
    case E_NUM_TYPE_COMP : sprintf(buff,"%g+I%g",
                                E_NUMBER_VAL_COMP(np).re, 
                                E_NUMBER_VAL_COMP(np).im); 
                           break;
  }
  
  E_NULL_ERROR(stp = e_string_new(0,buff,0));

  if(!(mode & E_TO_STR_NO_TAB))
  { /* add tab to the string */
    stp = e_string_insert_chars(stp,0,tab,' ');
  }

  return (stp);

error:
  if (stp) E_FREE(stp);
  return(0);
}

/*****************************************************************/
/* Clone a number                                                */
/* mode : cloning mode                                           */
/* If the number can't be cloned (default), return np            */
/* and increase its refs counter                                 */
/* np : a number, assumed to be of the correct type              */
/* Return : cloned e_number                                      */
/*          0 : Error                                            */
/* LINK : stp : 0                                                */
/*        RETURN : +1                                            */
/*****************************************************************/
void *e_number_clone(void *np,e_flag mode)
{ e_number *newnp = 0;

                 /* Can't clone : return the same object */
  if(E_OBJ_NO_CLONE(np)) return(E_USE(np));

  E_NULL_ERROR(newnp = e_number_new(0,E_NUMBER_TYPE(np),0));

                 /* copy the value */
  E_NUMBER_VAL(newnp) = E_NUMBER_VAL(np); 

  return(newnp);

error:
  if(newnp) E_FREE(newnp); 
  return(0);
}

/***************************************************************/
/* Test if the 2 numbers are equal:                            */
/* The two numbers are equal if they are identical (p1 == p2), */
/* or if they have the same type and value.                    */
/* WARNING : Two numbers of different types but with the same  */
/* value are considered as different.                          */
/* p1 is assumed to be of type e_number                        */
/* Return : 1 : p1 and p2 are equal                            */
/*          0 : p1 and p2 are different                        */
/* Link : p1 0                                                 */
/*        p2 0                                                 */
/***************************************************************/
int e_number_equals(void *p1,void *p2)
{ 

  if (p1 == p2)   return(1);
  if (!p1 || !p2) return(0);

            /* different objet  types => differ */
  if(E_OBJ_DESCR(p1) != E_OBJ_DESCR(p2)) return(0);

            /* different types => differ */
  if(E_NUMBER_TYPE(p1) != E_NUMBER_TYPE(p2)) return(0);

            /* then compare the values */

  switch(E_NUMBER_TYPE(p1))
  { case E_NUM_TYPE_CHAR : 
            return(E_NUMBER_VAL_CHAR(p1)==E_NUMBER_VAL_CHAR(p2));
    case E_NUM_TYPE_INT :  
            return(E_NUMBER_VAL_INT(p1)==E_NUMBER_VAL_INT(p2));
    case E_NUM_TYPE_REAL : 
            return(E_NUMBER_VAL_REAL(p1)==E_NUMBER_VAL_REAL(p2));
    case E_NUM_TYPE_COMP : 
             return((E_NUMBER_VAL_COMP(p1).re==E_NUMBER_VAL_COMP(p2).re) &&
                    (E_NUMBER_VAL_COMP(p1).im==E_NUMBER_VAL_COMP(p2).im)
                   );
  }
  return(0);

}
