#include <easy.h>
#include <stdlib.h>

e_obj_descriptor e_obj_descriptor_stack = {
  "stack",            /* Name      */
  0,                  /* def_mode  */
  malloc,             /* _alloc    */
  free,               /* _free     */
  realloc,            /* _realloc  */
  e_stack_free,       /* free      */
  e_stack_to_string,  /* to_string */
  e_stack_clone,      /* clone     */
  e_stack_equals,     /* equals    */
};


/***********************************************************/
/* STACK : typical usage                                   */
/***********************************************************/
/* A : dynamic allocation                                  */
/***********************************************************/
/* e_stack *sp;                                            */
/* char c;                                                 */
/*  if(!(sp = e_statck_(0,E_STACK_TYPE_CHAR))) goto error; */
/*  if(!e_stack_push(sp,'a')) goto error;                  */
/*  if(!e_stack_pop(sp,&c)) goto error;                    */
/*  e_stack_free(sp);                                      */
/***********************************************************/
/* B : static allocation                                   */
/***********************************************************/
/* e_stack ksp;                                             */
/* char c;                                                 */
/*  if(!e_stack_new(&sp,E_STACK_TYPE_CHAR)) goto error;    */
/*  if(!e_stack_push(&sp,'a')) goto error;                 */
/*  if(!e_stack_pop(&sp,&c)) goto error;                   */
/*  e_stack_free(&sp);                                     */
/***********************************************************/

/***************************************************************************/
/* Create or initialise a new kstatck                                      */
/* sp : e_stack to initialise (static stack)                               */
/*       if NULL allocate it                                               */
/*      sp is assumed to be of the correct type                            */
/* type : one of E_STACK_TYPE_CHAR, E_STACK_TYPE_INT, E_STACK_TYPE_POINTER */
/* size : inital size : if NULL use E_STACK_DEF_SIZE                       */
/* Return e_statck                                                         */
/*        0 : error                                                        */
/***************************************************************************/
e_stack *e_stack_new(e_stack *sp,int type,int size)
{ 
  if (!sp)
  { if (!(sp = e_obj_descriptor_stack._alloc(sizeof(e_stack)))) 
         return((e_stack *)e_error_no_memory("e_stack_new"));
    sp->h.refs = 1;
  }
  else sp->h.refs = -1;
  sp->h.mode = 0;

  sp->h.obj_desc_p = &e_obj_descriptor_stack;

  if (size <= 0) size = E_STACK_DEF_SIZE;

                   /* Allocate stack buffer */
  if (!(sp->buf = e_obj_descriptor_stack._alloc(
                              size*sizeof(e_stack_obj))))
  {                /* free sp if is was mallocated */
    if (sp->h.refs > 0) e_obj_descriptor_stack._free(sp);  
    e_error_no_memory("e_stack_new");
    goto error;
  }

  sp->type = type;
  sp->size = size;
  sp->index = 0;

  return(sp);

error:
  if(sp && (sp->h.refs > 0)) e_obj_descriptor_stack._free(sp);
  return(0);
}

/***********************************************/
/* Free e_stack and free it:                   */
/* Free stack buffer and set it to NULL        */
/* If sp was mallocated, free it and return 0 */
/***********************************************/
/***************************************************************************/
/* Free the stack                                                          */
/* If the refs counter is -1 : static stack : free the stack buffer only   */
/* If the refs counter is 1 : free the allocated memory for the e_stack    */
/* Otherwise decrement the counter by 1.                                   */
/* Link : stp : -1                                                         */
/***************************************************************************/
void e_stack_free(void *sp)
{

  if(sp)
  { if(E_OBJ_REF(sp) == 0)
    { e_error(E_ERROR_INCONSISTENCY,E_ERROR_LEVEL_PANIC,
            "Trying to free a free stack (%x) ",sp);
      goto error; 
    }

    if (E_OBJ_REF(sp) < 0)      /* static stack, free the buffer */
    { if (E_STACK_BUF(sp)) 
          e_obj_descriptor_stack._free(E_STACK_BUF(sp));
      E_STACK_BUF(sp) = 0;
      return;
    }

    if (E_OBJ(sp)->refs-- > 1) return;   /* decrement counter    */

                                /* Last reference, 
                                   free the stack buffer and the stack */
    if (E_STACK_BUF(sp)) e_obj_descriptor_stack._free(E_STACK_BUF(sp));
    e_obj_descriptor_string._free(sp);
  }
  return;

error:
  e_error_flush(E_ERROR_LEVEL_DEBUG); 
  
}

/********************************************************************/
/* Push the object pointed to by p on the stack                     */
/* p must match the type of object corresponding to the stack type. */
/* return : 1 OK                                                    */
/*          0 : error                                               */
/********************************************************************/
int e_stack_push(e_stack *sp,void *p)
{ void *bp;

  E_NULL_ERROR(e_check_type(sp,E_TYPE_STACK,"e_stack_push(sp,-)"));
  if(!p) return(e_error_null_pointer("e_stack_push(-,p)"));

                     /* If the buffer is too small, double its size */
  if (sp->index >= sp->size)
  { if(!(bp = e_obj_descriptor_stack._realloc(sp->buf,sp->size*2)))
         return(e_error_no_memory("e_stack_push"));
    sp->buf = bp;
    sp->size *= 2;
  }

  switch(sp->type)
  { case E_STACK_TYPE_CHAR : sp->buf[(sp->index)++].c = *((char *)p); 
                             break;
    case E_STACK_TYPE_INT :  sp->buf[(sp->index)++].i = *((int *)p); 
                             break;
    case E_STACK_TYPE_POINTER : sp->buf[(sp->index)++].p = *((void **)p); 
                             break;
  }
  return(1);

error:
  return(0);
}

/********************************************************************/
/* Pop the top stack object and write it to *p.                     */
/* p must match the type of object corresponding to the stack type. */
/* return : 1 OK                                                    */
/*          0 : error (empty stack or sp is NULL)                   */        
/********************************************************************/
int e_stack_pop(e_stack *sp,void *p)
{
  E_NULL_ERROR(e_check_type(sp,E_TYPE_STACK,"e_stack_pop(sp,-)"));
  if(!p) return(e_error_null_pointer("e_stack_pop(-,p)"));
  if(sp->index <= 0) return(0);

  switch(sp->type)
  { case E_STACK_TYPE_CHAR : *((char *)p) = sp->buf[--(sp->index)].c; 
                             break;
    case E_STACK_TYPE_INT : *((int *)p) = sp->buf[--(sp->index)].i; 
                            break;
    case E_STACK_TYPE_POINTER : *((void **)p) = sp->buf[--(sp->index)].p; 
                            break;
  }
  return(1);

error:
  return(0);
}

/*************************************************************/
/* Convert the stack to a string                             */
/* The stack content is printed out as a coma separated list */
/* starting at the bottom of the stack:                      */
/* Stack(TYPE,i1,i2,...iN)                                   */
/* sp : a stack, assumed to be of the correct type           */
/* Return : new string                                       */
/*          0 : error                                        */
/* LINK : sp : 0                                             */
/*        RETURN : +1                                        */
/*************************************************************/
void *e_stack_to_string(void *sp,int tab,e_flag mode)
{ e_string *stp = 0;
  e_string *stp2 = 0;
  char buff[E_BUFF_SIZE],c,*s;
  int i;
  e_stack_obj *op;

  E_NULL_ERROR(stp = e_string_new(0,"Stack(",E_STACK_INDEX(sp)*10));

                          /* type of stack */
  switch(E_STACK_TYPE(sp))
  { case E_STACK_TYPE_CHAR    : s = "\"Char\","; break;
    case E_STACK_TYPE_INT     : s = "\"Int\","; break;
    case E_STACK_TYPE_POINTER : s = "\"Pointer\","; break;
    case E_STACK_TYPE_OBJECT  : s = "\"Object\","; break;
    default                   : s = ""; break;
  }
  E_NULL_ERROR(stp = e_string_append_buffer(stp,s));

  op = E_STACK_BUF(sp);
  for(i=0; i < E_STACK_INDEX(sp); i++)
  {                       /* c , or ) for last item */
    c = (i ==(E_STACK_INDEX(sp)-1)) ?  ')': ',';  
    switch(E_STACK_TYPE(sp))
    { case E_STACK_TYPE_CHAR :
            sprintf(buff,"%c%c",op[i].c,c);
            E_NULL_ERROR(stp = e_string_append_buffer(stp,buff));
            break;
      case E_STACK_TYPE_INT :
            sprintf(buff,"%d%c",op[i].i,c);
            E_NULL_ERROR(stp = e_string_append_buffer(stp,buff));
            break;
      case E_STACK_TYPE_POINTER :
            sprintf(buff,"%x%c",(unsigned int)op[i].p,c);
            E_NULL_ERROR(stp = e_string_append_buffer(stp,buff));
            break;
      case E_STACK_TYPE_OBJECT :
            E_NULL_ERROR(stp2=E_TO_STRING(op[i].p,0,E_TO_STR_EMBEDED));
            E_NULL_ERROR(stp = e_string_append_string(stp,stp2));
            stp2 = 0;
            sprintf(buff,"%c",c);
            break;
      default :
            break;
    }
  }  

  if(!(mode & E_TO_STR_NO_TAB))
  { /* add tab to the string */
    stp = e_string_insert_chars(stp,0,tab,' ');
  }

  return (stp);

error:
  if (stp) E_FREE(stp);
  if (stp2) E_FREE(stp2);
  return(0);
}

/***************************************************/
/* Clone a stack                                   */
/* mode : cloning mode                             */
/* If the strack can't be cloned, return sp        */
/* and increase its refs counter                   */
/* sp : a stack, assumed to be of the correct type */
/* Return : cloned e_stack                         */
/*          0 : Error                              */
/* LINK : stp : 0                                  */
/*        RETURN : +1                              */
/***************************************************/
void *e_stack_clone(void *sp,e_flag mode)
{ e_stack *newsp;
  int i;

                 /* Can't clone : return the same object */
  if(E_OBJ_NO_CLONE(sp)) return(E_USE(sp));

  if((newsp = e_stack_new(0,E_STACK_TYPE(sp),E_STACK_SIZE(sp))))
  { E_STACK_INDEX(newsp) = E_STACK_INDEX(sp);

                 /* copy the stack buffer */
    for(i =0; i < E_STACK_INDEX(sp); i++)
    {            /* Copy static data or clone the objects */
      if(E_STACK_TYPE(sp) == E_STACK_TYPE_OBJECT)
      { E_STACK_BUF(newsp)[i].p = E_CLONE(E_STACK_BUF(newsp)[i].p,mode);
      }
      else E_STACK_BUF(newsp)[i] = E_STACK_BUF(sp)[i];
    }
  }
  return(newsp);

}

/******************************************************************/
/* Test if the 2 stacks are equals                                */
/* There is little use for this, but we test the indices and each */
/* Object on the stack                                            */
/* Return : 1 : p1 == p2                                          */
/*          1 : p1 and p2 are different                           */
/* REF : p1,p2 : 0                                                */
/******************************************************************/
int e_stack_equals(void *p1,void *p2)
{ int test,i;

  if (p1 == p2) return(1);
  
  if (!p1 || !p2) return(0);
 
            /* different types,  index  => differ */
  if((E_STACK_TYPE(p1)  != E_STACK_TYPE(p2)) ||
     (E_STACK_INDEX(p1) != E_STACK_INDEX(p2))) 
     return(0);

            /* compare all objects on the stack */         
  for(i=0;E_STACK_INDEX(p1); i++)
  { switch(E_STACK_TYPE(p1))
    { case E_STACK_TYPE_CHAR : 
           test = (E_STACK_BUF(p1)[i].c == E_STACK_BUF(p2)[i].c); 
           break;
	     case E_STACK_TYPE_INT :
           test = (E_STACK_BUF(p1)[i].i == E_STACK_BUF(p2)[i].i); 
           break;
      case E_STACK_TYPE_POINTER : 
           test = (E_STACK_BUF(p1)[i].p == E_STACK_BUF(p2)[i].p); 
           break;
      case E_STACK_TYPE_OBJECT : 
           test = E_EQUALS(E_STACK_BUF(p1)[i].p,E_STACK_BUF(p2)[i].p);
           break;
      default : test = 0; break;
    }
    if(!test) return (0);
  }


  return(1);

}
